/******************************************************************************
  (c) Copyright 2002-2006, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile: MetaToXML.java,v $
  Version:       $Name:  $ $Revision: 1.3 $
  Last Modified: $Date: 2008/05/19 13:36:48 $
 *****************************************************************************/
package org.netkernel.ext.system.transrepresentation;

import java.io.IOException;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.meta.IPhysicalEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.meta.IRepresentationMeta;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSPredicate;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardTransreptorImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
//import org.netkernel.xml.xda.IXDAReadOnly;
import org.w3c.dom.Document;


public class MetaToXML extends StandardTransreptorImpl
{
	public MetaToXML()
	{	this.declareThreadSafe();
		this.declareName("MetaToXML");
		this.declareDescription("serializes an IMetaRepresentation to XML");
		this.declareFromRepresentation(IMetaRepresentation.class);
		this.declareToRepresentation(String.class);
		this.declareToRepresentation(IReadableBinaryStreamRepresentation.class);
		this.declareToRepresentation(Document.class);
	}
	
	public void onTransrept(INKFRequestContext context) throws Exception
	{	
		IMetaRepresentation meta=(IMetaRepresentation)context.sourcePrimary(IMetaRepresentation.class);
		
		IRepDeployedModules dmr= (IRepDeployedModules)context.source("active:moduleList", IRepDeployedModules.class);
		
		StringWriter sw = new StringWriter(1024);
		sw.write("<meta>");
		writeMeta(meta,sw);
		
		if (meta instanceof IEndpointMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",IEndpointMeta.class.getName());
			writeEndpoint( (IEndpointMeta)meta, sw, dmr);
		}
		else if (meta instanceof IPrototypeMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",IPrototypeMeta.class.getName());
			writePrototype( (IPrototypeMeta)meta, sw,dmr);
		}
		else if (meta instanceof IRepresentationMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",IRepresentationMeta.class.getName());
			writeRepresentation( (IRepresentationMeta)meta, sw, dmr);
		}
		else if (meta instanceof ISpaceMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",ISpaceMeta.class.getName());
			writeSpace( (ISpaceMeta)meta, sw, dmr);
		}
		else if (meta instanceof IModuleMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",IModuleMeta.class.getName());
			writeModule( (IModuleMeta)meta, sw,dmr, context);
		}
		else if (meta instanceof IPhysicalEndpointMeta)
		{	
			XMLUtils.writeEscaped(sw,"class",IPhysicalEndpointMeta.class.getName());
			writePhysicalEndpoint( (IPhysicalEndpointMeta)meta, sw,dmr,context);
		}
		
		
		sw.write("</meta>");
		
		Object rep=sw.toString();
		Class toClass=context.getThisRequest().getRepresentationClass();
		if (!toClass.isAssignableFrom(String.class))
		{	rep=context.transrept(rep, toClass);
		}
		
		INKFResponse resp = context.createResponseFrom(rep);
		resp.setMimeType("text/xml");
	}
	
	private void writeMeta(IMetaRepresentation meta, StringWriter sw) throws IOException
	{
		String name = meta.getName();
		if (name!=null && name.length()>0)
		{	XMLUtils.writeEscaped(sw,"name",name);
		}
		String desc = meta.getDescription();
		if (desc!=null && desc.length()>0)
		{	XMLUtils.writeEscaped(sw,"desc",desc);
		}
		IIdentifier icon = meta.getIcon();
		if (icon!=null)
		{	XMLUtils.writeEscaped(sw,"icon",icon.toString());
		}
		IIdentifier hrd = meta.getHumanReadableDocumentation();
		if (hrd!=null)
		{	XMLUtils.writeEscaped(sw,"hrd",hrd.toString());
		}
		IRequestResponseFields fields=meta.getAdditionalFields();
		sw.append("<fields>");
		for (int i=0; i<fields.size(); i++)
		{
			sw.write("<field name='");
			sw.write(fields.getKey(i));
			sw.write("'>");
			String value;
			try
			{	value=fields.getValue(i).toString();
			}
			catch (Exception e)
			{	value="";
			}
			sw.write(XMLUtils.escape(value));
			sw.write("</field>");
		}
		sw.append("</fields>");
	}
	
	private void writeEndpoint(IEndpointMeta meta, StringWriter sw, IRepDeployedModules aDMR) throws IOException
	{
		sw.write("<interface>");
		sw.write("<verbs>");
		int verbs = meta.getSupportedVerbs();
		int mask=1;
		while (verbs>0)
		{	if ( (verbs&mask) >0 )
			{	XMLUtils.write(sw,"verb",RequestFactory.verbString(mask));
				verbs^=mask;
			}
			mask<<=1;
		}
		sw.write("</verbs>");
	
		if (meta.getIdentifierGrammar()!=null)
		{
			XMLUtils.write(sw,"grammar",meta.getIdentifierGrammar().toXMLString());
		}
		IEndpointArgumentMeta args[] = meta.getArguments();
		sw.write("<args>");
		for (int i=0; i<args.length; i++)
		{	IEndpointArgumentMeta argMeta = args[i];
			sw.write("<arg>");
			writeMeta(argMeta, sw);
			XMLUtils.writeEscaped(sw,"isId",Boolean.toString(argMeta.isIdentifier()));
			Class[] reqRep = argMeta.getRequiredRepresentations();
			sw.write("<reps>");
			for (int j=0; reqRep!=null && j<reqRep.length; j++)
			{	sw.write("<rep>");
				writeClass(reqRep[j],sw,aDMR);
				sw.write("</rep>");
			}
			sw.write("</reps>");
			sw.write("</arg>");
		}
		sw.write("</args>");
		
		IMetaRepresentation[] exs = meta.getExceptions();
		sw.write("<exs>");
		for (int i=0; i<exs.length; i++)
		{	sw.write("<ex>");
			writeMeta(exs[i], sw);
			sw.write("</ex>");
		}
		sw.write("</exs>");
		
		Class[] respClasses = meta.getResponseClasses();
		sw.write("<reps>");
		for (int j=0; respClasses!=null && j<respClasses.length; j++)
		{	Class c = respClasses[j];
			sw.write("<rep>");
			writeClass(c,sw,aDMR);
			sw.write("</rep>");
		}
		sw.write("</reps>");
		
		
		
		sw.write("</interface>");
	}
	
	private void writeClass(Class aClass, StringWriter sw, IRepDeployedModules aDMR) throws IOException
	{
		String hash=aDMR.hashForClass(aClass);
		if (hash!=null)
		{	XMLUtils.writeEscaped(sw,"hash", hash);
			IMetaRepresentation cm=aDMR.metadataForHash(hash);
			XMLUtils.writeEscaped(sw,"name", cm.getName());
		}
		XMLUtils.writeEscaped(sw,"class", aClass.getName().toString());	
	}
	
	private void writePrototype(IPrototypeMeta meta, StringWriter sw, IRepDeployedModules aDMR) throws IOException
	{
		IPrototypeParameterMeta[] parameters = meta.getParameters();
		XMLUtils.write(sw,"implementation",meta.getImplementation().getName());
		XMLUtils.write(sw,"identifier",meta.getIdentifier().toString());
		
		sw.write("<parameters>");
		for (int i=0; i<parameters.length; i++)
		{	IPrototypeParameterMeta p=parameters[i];
			sw.write("<parameter>");
			writeMeta(p,sw);
			XMLUtils.write(sw,"class",p.getRepresentationalClass().getName());
			String v;
			if (p.getDefault()==null)
			{	v="null";
			}
			else
			{	v=p.getDefault().toString();
			}
			XMLUtils.write(sw,"default",v);
			XMLUtils.write(sw,"min",Integer.toString(p.getMinimumValency()));
			XMLUtils.write(sw,"max",Integer.toString(p.getMaximumValency()));
			
			sw.write("</parameter>");
		}
		sw.write("</parameters>");
	}
	
	private void writeRepresentation(IRepresentationMeta meta, StringWriter sw, IRepDeployedModules aDMR) throws IOException
	{
		Class impl = meta.getImplementation();
		XMLUtils.write(sw,"implementation",impl.getName());
		Method[] methods = impl.getMethods();
		sw.write("<methods>");
		for (int i=0; i<methods.length; i++)
		{	Method m = methods[i];
			if (m.getDeclaringClass()!=java.lang.Object.class)
			{
				sw.write("<method>");
				XMLUtils.write(sw,"name",m.getName());
				XMLUtils.write(sw,"return",m.getReturnType().getName());
				Class[] exceptions=m.getExceptionTypes();
				writeClassList("exceptions","exception",sw,exceptions);
				Class[] params=m.getParameterTypes();
				writeClassList("params","param",sw,params);
				sw.write("</method>");
			}
		}
		sw.write("</methods>");
	}
	
	private void writeClassList(String aCollection, String aItem, StringWriter sw, Class[] aClasses) throws IOException
	{
		sw.write('<');
		sw.write(aCollection);
		sw.write('>');
		
		for (int i=0; i<aClasses.length; i++)
		{	Class c = aClasses[i];
			XMLUtils.writeEscaped(sw, aItem, c.getName());
		}
		
		sw.write("</");
		sw.write(aCollection);
		sw.write('>');
	}
	
	private void writeSpace(ISpaceMeta meta, StringWriter sw, IRepDeployedModules aDMR) throws IOException
	{
		String ourHash=aDMR.hashForMetadata(meta);
		ISpace ourCtx=(ISpace)aDMR.componentForHash(ourHash);
		boolean publicContext=false;
		if (ourCtx instanceof ISpaceWithIdentity)
		{	ISpaceWithIdentity ourCtxIdentity=(ISpaceWithIdentity)ourCtx;
			publicContext=getKernel().getSpace(ourCtxIdentity.getIdentifier(), ourCtxIdentity.getVersion(), ourCtxIdentity.getVersion())!=null;
			XMLUtils.writeEscaped(sw,"identifier",ourCtxIdentity.getIdentifier().toString());
			XMLUtils.writeEscaped(sw,"version",ourCtxIdentity.getVersion().toString());
		}
		XMLUtils.writeEscaped(sw,"public",Boolean.toString(publicContext));
		
		List<IEndpoint> physicalEndpoints = new ArrayList<IEndpoint>();
		ISpaceElements parts = meta.getElements();
		int s=parts.size();
		for (int i=0; i<s; i++)
		{	IEndpoint physicalEndpoint = parts.getPhysicalEndpoint(i);
			if (physicalEndpoint!=null)
			{	if (!physicalEndpoints.contains(physicalEndpoint))
				{	physicalEndpoints.add(physicalEndpoint);
				}
			}			
		}
		
		sw.write("<endpoints>");
		for (Iterator<IEndpoint> j=physicalEndpoints.iterator(); j.hasNext(); )
		{	IEndpoint ep = j.next();
			String hash=aDMR.hashForComponent(ep);
			sw.write("<endpoint>");
			XMLUtils.writeEscaped(sw,"hash",hash);
			if (hash.equals("cn/t/Vh"))
			{	System.out.println("here");
			}
			if (hasProblems(hash, aDMR))
			{	sw.write("<problem/>");
			}
			
			// write out all elements for this mapping
			for (int i=0; i<s; i++)
			{
				if (parts.getPhysicalEndpoint(i)==ep)
				{
					IIdentifier partId=parts.getIdentifier(i);
					if (partId!=null)
					{
						sw.write("<element>");
						String id=partId.toString();
						XMLUtils.writeEscaped(sw,"id",id);
						String partHash=ourHash+"/"+id;
						XMLUtils.writeEscaped(sw,"hash",partHash);
						IMetaRepresentation partMeta = aDMR.metadataForHash(partHash);
						if (partMeta!=null)
						{	writeMeta(partMeta, sw);
						}
						
						ISpace delegate = parts.getDelegatedSpace(i);
						boolean local=delegate==null || delegate==ourCtx;
						XMLUtils.write(sw,"local",Boolean.toString(local));
						String delegateHash=aDMR.hashForComponent(delegate);
						if (delegateHash!=null)
						{	XMLUtils.writeEscaped(sw,"delegate-hash",delegateHash);
							IMetaRepresentation delegateMeta=aDMR.metadataForHash(delegateHash);
							if (delegateMeta!=null)
							{	XMLUtils.writeEscaped(sw,"delegate-name",delegateMeta.getName());
							}
						}
						
						String type = spaceElementMetaClassToType(partMeta);
						XMLUtils.write(sw,"type",type);
						
						if (hasProblems(partHash, aDMR))
						{	sw.write("<problem/>");
						}
						
						
						sw.write("</element>");
					}
				}
			}
			
			XMLUtils.writeEscaped(sw,"name",ep.toString());
			XMLUtils.writeEscaped(sw,"class",Utils.shortClassName(ep.getClass()));
			
			sw.write("</endpoint>");
		}
		sw.write("</endpoints>");
		
	}
	
	private void writeModule(IModuleMeta meta, StringWriter sw, IRepDeployedModules aDMR, INKFRequestContext aContext) throws IOException
	{	XMLUtils.writeEscaped(sw, "source", meta.getSource().toString());
		XMLUtils.writeEscaped(sw, "identifier", meta.getIdentifier().toString());
		XMLUtils.writeEscaped(sw, "version", meta.getVersion().toString());

		sw.write("<spaces>");
		String thisModuleHash = aDMR.hashForMetadata(meta);
		if (thisModuleHash!=null)
		{	IModule thisModule = (IModule)aDMR.componentForHash(thisModuleHash);
			if (thisModule!=null)
			{	
				ISpace[] spaces=thisModule.getSpaces();
				//ISpace[] publicSpaces=thisModule.getPublicSpaces();
				for (int i=0; i<spaces.length; i++)
				{	ISpace space=spaces[i];
				
					writeModuleSpace(space,sw,aDMR,aContext);
				
					
					
				}
			}
		}
		sw.write("</spaces>");
		
	}
	
	private void writeModuleSpace(ISpace aSpace, StringWriter sw, IRepDeployedModules aDMR, INKFRequestContext aContext) throws IOException
	{
		sw.write("<space>");
		String spaceHash=aDMR.hashForComponent(aSpace);
		
		boolean isPublic=false;
		if (aSpace instanceof ISpaceWithIdentity)
		{	ISpaceWithIdentity spi=(ISpaceWithIdentity)aSpace;
			if (spi.getIdentifier()!=null)
			{	ISpaceWithIdentity spaceFromKernel=getKernel().getSpace(spi.getIdentifier(), spi.getVersion(), spi.getVersion());
				isPublic=(spaceFromKernel!=null);
				XMLUtils.writeEscaped(sw,"id",spi.getIdentifier().toString());
			}
		}
		
		ISpaceMeta spaceMeta=(ISpaceMeta)aDMR.metadataForHash(spaceHash);
		XMLUtils.writeEscaped(sw,"name",spaceMeta.getName());
		XMLUtils.writeEscaped(sw,"public",Boolean.toString(isPublic));
		XMLUtils.writeEscaped(sw,"hash",spaceHash);
		if (hasProblems(spaceHash, aDMR))
		{	sw.append("<problem/>");
		}
		
		//nested spaces
		sw.append("<spaces>");
		ISpaceElements elements=spaceMeta.getElements();
		List<IEndpoint> physicalEndpoints = new ArrayList<IEndpoint>();					
		for (int j=0; j<elements.size(); j++)
		{	IEndpoint pe = elements.getPhysicalEndpoint(j);
			if (!physicalEndpoints.contains(pe))
			{	physicalEndpoints.add(pe);
				if (pe instanceof IStandardEndpoint)
				{	IStandardEndpoint spe = (IStandardEndpoint)pe;
					try
					{	ISpace[] subspaces=spe.getSpaces(aContext);
						for (int k=0; k<subspaces.length; k++)
						{	ISpace subspace=subspaces[k];
							writeModuleSpace(subspace, sw,aDMR,aContext);
						}
					}
					catch (Exception e)
					{	System.out.println(Utils.throwableToString(e));
					}
					
				}
			
			}
		}
		sw.append("</spaces>");
		sw.write("</space>");
	}
	
	
	private void writePhysicalEndpoint(IPhysicalEndpointMeta meta, StringWriter sw,  IRepDeployedModules aDMR, INKFRequestContext context) throws IOException
	{
		Class impl = meta.getImplementation();
		XMLUtils.write(sw,"implementation",impl.getName());

		//write logical endpoints
		String hash=aDMR.hashForMetadata(meta);
		if (hash!=null)
		{
			IEndpoint ep = (IEndpoint)aDMR.componentForHash(hash);
			int j=hash.lastIndexOf('/');
			String spaceHash=hash.substring(0,j);
			ISpace parentSpace = (ISpace)aDMR.componentForHash(spaceHash);
			ISpaceMeta parentSpaceMeta=(ISpaceMeta)aDMR.metadataForHash(spaceHash);
			ISpaceElements parts=parentSpaceMeta.getElements();
			int size=parts.size();
			sw.write("<elements>");
			for (int i=0; i<size; i++)
			{	if (parts.getPhysicalEndpoint(i)==ep)
				{	
					IIdentifier partId=parts.getIdentifier(i);
					if (partId!=null)
					{
						sw.write("<element>");
						String partHash=spaceHash+"/"+partId.toString();
						XMLUtils.write(sw,"hash",partHash);
						XMLUtils.write(sw,"id",partId.toString());
						IMetaRepresentation partMeta = aDMR.metadataForHash(partHash);
						if (partMeta!=null)
						{	XMLUtils.write(sw,"name",partMeta.getName());
							XMLUtils.write(sw,"type",spaceElementMetaClassToType(partMeta));
						}
						ISpace delegate = parts.getDelegatedSpace(i);
						boolean local=delegate==null || delegate==parentSpace;
						XMLUtils.write(sw,"local",Boolean.toString(local));
						String delegateHash=aDMR.hashForComponent(delegate);
						if (delegateHash!=null)
						{	XMLUtils.writeEscaped(sw,"delegate-hash",delegateHash);
							IMetaRepresentation delegateMeta=aDMR.metadataForHash(delegateHash);
							if(delegateMeta!=null)
							{	XMLUtils.writeEscaped(sw,"delegate-name",delegateMeta.getName());
							}
						}				
						
						if (hasProblems(partHash, aDMR))
						{	sw.write("<problem/>");
						}
						
						sw.write("</element>");
					}
				}
			}
			sw.write("</elements>");
		}
	}
	
	public static String spaceElementMetaClassToType(IMetaRepresentation aMeta)
	{
		String type;
		if (aMeta instanceof IEndpointMeta)
		{	type="endpoint";
		}
		else if (aMeta instanceof IPrototypeMeta)
		{	type="prototype";
		}
		else if (aMeta instanceof IRepresentationMeta)
		{	type="representation";
		}
		else
		{	type="??";
		}
		return type;
	}
	
	private boolean hasProblems(String aHash, IRepDeployedModules aDMR)
	{	IMetaRepresentation meta= aDMR.metadataForHash(aHash);
		boolean result=(meta==null);
		if (!result)
		{	result=meta.getAdditionalFields().getValue(IEndpointStateMeta.KEY_PROBLEM)!=null;
		}
		return result;
	}
	
}
